#
#   MLCV Summer 2016
#   Homework Assignment 04 - Gaussian Mixture Models
#
#   Exercise 2) K-Means and EM algorithm
#
###########################################################################

import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np
from time import time
from scipy.spatial.distance import cdist

def compute_assignment(X, means, distance='sqeuclidean'):
    print('Computing distances...')
    n, d = X.shape

    t = time()
    D = cdist(X, means, distance)
    assignment = np.argmin(D, axis=1)
    assignment_cost = np.sum(D[np.arange(n), assignment])

    t = time() - t
    print('Datapoints assigned. Took %fs, cost=%f' % (t, assignment_cost))
    return assignment, assignment_cost


###########################################################################
def cond_plot(it, title=None, data=None, colors=None, marker='o', ax=None):

    plt.sca(ax)
    plt.scatter(data[:, 0], data[:, 1], c=colors, marker=marker, lw=0, s=50)

    if title is not None:
        plt.title('%s - %s' % (it, title), fontweight='bold')
    plt.draw()
    plt.pause(0.5)



def mkmeans( X, K, ax1=None, ax2=None):
    #MKMEANS Summary of this function goes here
    # X - nxdim matrix, the data: n samples with dim properties each
    # K - number of clusters to assign the data to
    # ax1 - axis to plot the cost over time
    # ax2 - axis to plot the mean-shift over time
    n, d = X.shape

    # Initialize cluster means randomly within the data range
    Xmin = np.min(X, axis=0)
    Xmax = np.max(X, axis=0)
    Xrange = Xmax - Xmin
    means = np.array([Xmin + np.random.rand(d)*Xrange for _ in range(K)])

    if ax2 is not None:
        ax2.set_title('Mean shift')

    if ax1 is not None:
        ax1.set_title('K-means cost')

    if ax1 is not None or ax2 is not None:
        cmap = cm.get_cmap('jet')

    # Convergence criteria
    eTol = 1e-2
    converged = False
    maxIter = 100
    iteration = 0
    old_cost, old_meanshift = np.infty, 0
    while not converged and iteration < maxIter:

        print('Iteration %d' % iteration)
        assignment, cost = compute_assignment(X, means)
        # Re-initialize means if a cluster has no samples assigned to it
        while len(set(assignment)) != K:
            print('Some clusters are empty, re-initializing means...')
            means = np.array([Xmin + np.random.rand(d) * Xrange for _ in range(K)])
            assignment, cost = compute_assignment(X, means)

        # Compute new means
        old_means = means.copy()
        print('Computing new means...')
        for k in range(K):
            idx = np.array(assignment == k)
            means[k, :] = np.mean(X[idx, :], axis=0)
        print('Done.')

        # Compute the mean shift for each cluster
        print('Computing how much the means shifted...')
        sq_mean_shift = np.zeros(K)
        for k in range(K):
            diff = means[k, :] - old_means[k, :]
            sq_mean_shift[k] = np.dot(diff, diff)
            meanshift = np.linalg.norm(sq_mean_shift)
        print('Done.\n')

        # Visualize the cost
        if ax1 is not None and iteration > 0:

            dt = [iteration-1, iteration]
            ax1.plot(dt, [old_cost, cost], 'b-o')
            ax2.plot(dt, [old_meanshift, meanshift], 'b-o')

            plt.gcf().suptitle('Iter %d' % iteration)
            plt.draw()
            plt.pause(0.1)

        old_cost = cost
        old_meanshift = meanshift
        iteration += 1
        # Evaluate converge criterion
        if all([m < eTol for m in sq_mean_shift]):
            converged = True
            print('Converged after ', iteration, ' iterations')

    if not converged:
        print('Did not converge after ', iteration, ' iterations')

    return assignment
